@@ -4,7 +4,7 @@ package JSON::Schema::Modern::Vocabulary;
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Base role for JSON Schema vocabulary classes
 
-our $VERSION = '0.517';
+our $VERSION = '0.518';
 
 use 5.016;
 no if "$]" >= 5.031009, feature => 'indirect';
@@ -12,9 +12,12 @@ no if "$]" >= 5.033001, feature => 'multidimensional';
 no if "$]" >= 5.033006, feature => 'bareword_filehandles';
 use strictures 2;
 use JSON::Schema::Modern::Utilities qw(jsonp assert_keyword_type);
+use Carp ();
 use Moo::Role;
 use namespace::clean;
 
+our @CARP_NOT = qw(JSON::Schema::Modern);
+
 requires qw(vocabulary keywords);
 
 sub traverse {
@@ -84,7 +87,7 @@ JSON::Schema::Modern::Vocabulary - Base role for JSON Schema vocabulary classes
 
 =head1 VERSION
 
-version 0.517
+version 0.518
 
 =head1 SYNOPSIS
 
